MODULE bnf_chk
(*
	bnf_chk - EBNF syntax checker and formatter
	Copyright 2006 by icosaedro.it di Umberto Salsi <salsi@icosaedro.it>
	Info: http://www.icosaedro.it/bnf_chk/index.html
*)

IMPORT m2, io, str

CONST

	VERSION = "1.4"
	DEBUG = FALSE

VAR
	fn: STRING

	(* Preferences: *)
	do_report: BOOLEAN
	do_print_text: BOOLEAN
	do_print_html: BOOLEAN
	do_sort: BOOLEAN
	do_print_index: BOOLEAN
	tab_size: INTEGER
	print_file_name: BOOLEAN
	print_source: BOOLEAN
	print_context: BOOLEAN

	(* Warning messages already given: *)
	warn_invalid_id,  # using <ID> in place of ID
	warn_invalid_rule_terminator,  # using . in place of ;
	warn_invalid_eq, # using := or ::= in place of =
	warn_invalid_unquoted_lit_str: BOOLEAN

TYPE
	Identifier = RECORD
		name: STRING
		defined: BOOLEAN
		defined_in_line: INTEGER
		index: INTEGER
		used: INTEGER
	END

VAR
	ids: ARRAY OF Identifier


TYPE
	Node = RECORD
		type: (
			box,       # rule name - ex. A
			roundbox,  # literal string - ex. "PRINT"
			prod,      # product of factors - ex. A B C
			or,        # alternatives terms - ex. A|B|C
			repeat,    # zero or more times - ex. {A}
			opt        # optional - ex. [A]
		)
		id: Identifier  (* box *)
		s: STRING  (* roundbox *)
		n: Node  (* repeat, opt *)
		a: ARRAY OF Node  (* prod, or *)
	END

	Decl = RECORD      # a declaration, aka rule
		id: Identifier
		n: Node
	END

VAR
	decls: ARRAY OF Decl


(*
	SCANNER
*)

TYPE SYMBOL = (sym_eof, sym_semicolon,
	sym_name, sym_eq, sym_lbrace, sym_rbrace, sym_lsquare,
	sym_rsquare, sym_lround, sym_rround, sym_vbar, sym_string, sym_range)

VAR
	fd: FILE
	line: STRING
	(* current line being parsed *)

	line_n: INTEGER
	(* current line no. First line is the no. 1 *)

	line_idx: INTEGER
	(* offset of the current char in line *)

	line_pos: INTEGER
	(* text editor idea of the position of the current char *)

	c: STRING
	(* current char to be parsed, or NIL if end of the file *)

	sym: SYMBOL
	(* current symbol *)

	s: STRING
	(* ID or literal string, depending on the value of 's' *)


FUNCTION String2HTML(s: STRING): STRING
BEGIN
	s = substitute(s, "&", "&amp;")
	s = substitute(s, "<", "&lt;")
	s = substitute(s, ">", "&gt;")
	s = substitute(s, "\n", "<br>\n")
	RETURN s
END


FUNCTION print_text(s: STRING)
BEGIN
	IF do_print_html THEN
		s = String2HTML(s)
		s = substitute(s, "\t", "        ") # FIXME: missing col space count
		s = substitute(s, " ", "&nbsp;")
		s = "<code>" + s + "</code>"
	END
	print(s)
END


FUNCTION PrintLineSource()
BEGIN
	print_text( itos(line_n) )
	print_text( ":\t" )
	print_text( line )
	print_text( "\n" )
END


FUNCTION read_ch()
BEGIN
	# update line_n, line_pos:
	IF c = NIL THEN
	ELSIF c = "\t" THEN
		inc(line_pos, tab_size)
	ELSIF c = "\n" THEN
		inc(line_n, 1)
		line_pos = 1
	ELSE
		inc(line_pos, 1)
	END

	# get next char 'c' and update line_idx:
	IF line = NIL THEN
		c = NIL
	ELSIF line_idx < length(line) THEN
		c = line[line_idx]
		inc(line_idx, 1)
	ELSIF line_idx = length(line) THEN
		c = "\n"
		inc(line_idx, 1)
	ELSE
		TRY line = ReadLine(fd) END
		IF line = NIL THEN
			c = NIL
		ELSIF length(line) = 0 THEN
			c = "\n"
		ELSE
			c = line[0]
		END
		line_idx = 1
		IF print_source AND (line <> NIL) THEN
			PrintLineSource()
		END
	END

	IF DEBUG THEN
		print(c)
	END
END


FUNCTION message(msg_type: STRING, s: STRING)
BEGIN
	IF print_context THEN
		print_text("\n\t" + line + "\n\t" + str.repeat(" ", line_pos-1) + "\\_ HERE\n")
	END

	IF print_file_name THEN
		print_text(fn + ": ")
	END

	print_text("" + line_n + ":" + line_pos + ": ")

	IF do_print_html THEN
		print("<b>")
	END
	print_text(msg_type + ": " + s + "\n")
	IF do_print_html THEN
		print("</b>")
	END
END


(****
FUNCTION notice(s: STRING)
BEGIN
	message("notice", s)
END
****)


FUNCTION warning(s: STRING)
BEGIN
	message("Warning", s)
END


FUNCTION err(s: STRING)
BEGIN
	message("ERROR", s)
END


FUNCTION fatal(s: STRING)
BEGIN
	message("FATAL", s)
	exit(1)
END


FUNCTION read_lit_string(): STRING

	FUNCTION hex(c: STRING): INTEGER
	BEGIN
		IF (c >= "0") AND (c <= "9") THEN
			RETURN stoi(c)
		ELSIF (c >= "A") AND (c <= "F") THEN
			RETURN ASC(c) - ASC("A") + 10
		ELSIF (c >= "a") AND (c <= "f") THEN
			RETURN ASC(c) - ASC("a") + 10
		ELSE
			fatal("invalid hexadecimal digit `" + c + "'")
		END
	END

VAR  s: STRING  d1, d2: INTEGER
BEGIN
	s = "\""
	read_ch()
	LOOP
		IF c = NIL THEN
			fatal("unclosed literal string")

		ELSIF c = "\"" THEN
			read_ch()
			IF length(s) = 1 THEN
				fatal("invalid empty string")
			END
			RETURN s + "\""

		ELSIF c = "\\" THEN
			s = s + "\\"
			read_ch()
			IF (c = "\"") OR (c = "\\") THEN
				s = s + c
				read_ch()
			ELSIF c = "a" THEN
				s = s + "a"
				read_ch()
			ELSIF c = "b" THEN
				s = s + "b"
				read_ch()
			ELSIF c = "n" THEN
				s = s + "n"
				read_ch()
			ELSIF c = "r" THEN
				s = s + "r"
				read_ch()
			ELSIF c = "t" THEN
				s = s + "t"
				read_ch()
			ELSIF c = "x" THEN
				s = s + "x"
				read_ch()
				d1 = hex(c)
				s = s + c
				read_ch()
				d2 = hex(c)
				s = s + c
				read_ch()
				#s = s + CHR(16*d1+d2)
			ELSE
				fatal("invalid escape sequence \"\\" + c + "\"")
			END

		ELSIF (c >= " ") AND (ASC(c) <> 127) THEN
			s = s + c
			read_ch()

		ELSE
			err("invalid control character code " + ASC(c) + " in literal string -- possible unclosed string")
			read_ch()
			RETURN s + "\""

		END
	END
END


FUNCTION read_sym()

	FUNCTION WarnInvalidEq(s: STRING)
	BEGIN
		IF NOT warn_invalid_eq THEN
			warn_invalid_eq = TRUE
			warning("invalid `" + s + "' symbol -- please use `=' instead")
		END
	END

BEGIN
	WHILE (c = " ") OR (c = "\t") OR (c = "#") OR (c = "\n") OR (c = "\r") DO
		IF c = "#" THEN
			LOOP
				read_ch()
				IF (c = "\n") OR (c = NIL) THEN
					EXIT
				END
			END
		ELSE
			read_ch()
		END
	END
	IF c = NIL THEN
		sym = sym_eof
	ELSIF match(c, "[a-zA-Z_]") THEN
		s = c
		LOOP
			read_ch()
			IF match(c, "[a-zA-Z0-9_]") THEN
				s = s + c
			ELSE
				EXIT
			END
		END
		IF warn_invalid_id THEN
			IF NOT warn_invalid_unquoted_lit_str THEN
				warn_invalid_unquoted_lit_str = TRUE
				warning("guessed usage of the syntax style <ID> for IDs and STRING for literal strings -- trying to continue anyway")
			END
			s = "\"" + s + "\""  # FIXME: special chars must be escaped
			sym = sym_string
		ELSE
			sym = sym_name
		END
	ELSIF c = "<" THEN
		s = ""
		LOOP
			read_ch()
			IF match(c, "[\\-a-zA-Z0-9_ ]") THEN
				IF c = " " THEN
					c = "_"
				ELSIF c = "-" THEN
					c = "_"
				END
				s = s + c
			ELSIF c = ">" THEN
				IF NOT warn_invalid_id THEN
					warn_invalid_id = TRUE
					warning("invalid identifier syntax `<SOME ID>' -- please use `SOME_ID' instead")
				END
				read_ch()
				EXIT
			ELSE
				fatal("invalid char in <identifier>")
			END
		END
		sym = sym_name
	ELSIF c = "=" THEN
		sym = sym_eq
		read_ch()
	ELSIF c = ":" THEN
		read_ch()
		IF c = "=" THEN
			WarnInvalidEq(":=")
			read_ch()
			sym = sym_eq
		ELSIF c = ":" THEN
			read_ch()
			IF c = "=" THEN
				WarnInvalidEq("::=")
				read_ch()
				sym = sym_eq
			ELSE
				fatal("invalid character or symbol")
			END
		ELSE
			fatal("invalid character or symbol")
		END
	ELSIF c = "|" THEN
		sym = sym_vbar
		read_ch()
	ELSIF c = "{" THEN
		sym = sym_lbrace
		read_ch()
	ELSIF c = "}" THEN
		sym = sym_rbrace
		read_ch()
	ELSIF c = "[" THEN
		sym = sym_lsquare
		read_ch()
	ELSIF c = "]" THEN
		sym = sym_rsquare
		read_ch()
	ELSIF c = "(" THEN
		sym = sym_lround
		read_ch()
	ELSIF c = ")" THEN
		sym = sym_rround
		read_ch()
	ELSIF c = ";" THEN
		sym = sym_semicolon
		read_ch()
	ELSIF c = "\"" THEN
		s = read_lit_string()
		sym = sym_string
	ELSIF c = "." THEN
		read_ch()
		IF c = "." THEN
			read_ch()
			sym = sym_range
		ELSE
			IF NOT warn_invalid_rule_terminator THEN
				warn_invalid_rule_terminator = TRUE
				warning("invalid rule terminator `.' -- please use `;' instead")
			END
			sym = sym_semicolon
			read_ch()
		END
	ELSIF c > " " THEN
		err("unexpected character `" + c + "' -- trying to continue anyway")
		s = c
		read_ch()
		WHILE c > " " DO
			s = s + c
			read_ch()
		END
		s = "\"" + s + "\"" # FIXME: special chars must be escaped
		sym = sym_string
	ELSE
		err("unexpected control char -- trying to continue anyway")
		sym = sym_string
	END
END


FUNCTION add_id(name: STRING, decl: BOOLEAN): Identifier
VAR
	i: INTEGER
	id: Identifier
BEGIN
	FOR i=0 TO count(ids)-1 DO
		IF ids[i][name] = name THEN
			id = ids[i]
		END
	END
	IF id = NIL THEN
		IF decl THEN
			id[name] = name
			id[defined] = TRUE
			id[defined_in_line] = line_n
			id[used] = 0
		ELSE
			id[name] = name
			id[defined] = FALSE
			id[defined_in_line] = 0
			id[used] = id[used] + 1
		END
		ids[count(ids)] = id
	ELSE
		IF decl THEN
			IF id[defined] THEN
				fatal("id `" + name + "' already defined in line "
					+ itos(id[defined_in_line]))
			ELSE
				(* Dichiarazione di un ID gia' usato prima: *)
				id[defined] = TRUE
				id[defined_in_line] = line_n
			END
		ELSE
			id[used] = id[used] + 1
		END
	END
	RETURN id
END


FUNCTION ids_report()
VAR i: INTEGER   id: Identifier
BEGIN
	print_text("IDs report:\n")
	FOR i=0 TO count(ids)-1 DO
		id = ids[i]
		IF NOT id[defined] OR (id[used] = 0) THEN
			print_text(id[name] + ": ")
			IF id[defined] THEN
				print_text("defined in line " + itos(id[defined_in_line]) + ", ")
			ELSE
				print_text("NOT DEFINED, ")
			END
			IF id[used] > 0 THEN
				print_text("used " + itos(id[used]) + " times\n")
			ELSE
				print_text("NOT USED\n")
			END
		END
	END
END


(*
	PARSER
*)


FUNCTION read_expr(): Node


	FUNCTION read_factor(): Node
	VAR f: Node
	BEGIN
		IF sym = sym_lround THEN
			read_sym()
			f = read_expr()
			IF sym <> sym_rround THEN
				fatal("expected `)'")
			END
			read_sym()
		ELSIF sym = sym_lsquare THEN
			f[type] = opt
			read_sym()
			f[n] = read_expr()
			IF sym <> sym_rsquare THEN
				fatal("expected `]'")
			END
			read_sym()
		ELSIF sym = sym_lbrace THEN
			f[type] = repeat
			read_sym()
			f[n] = read_expr()
			IF sym <> sym_rbrace THEN
				fatal("expected `}'")
			END
			read_sym()
		ELSIF sym = sym_name THEN
			f[type] = box
			f[id] = add_id(s, FALSE)
			read_sym()
		ELSIF sym = sym_string THEN
			f[type] = roundbox
			##f[s] = StringToLiteral(s)
			f[s] = s
			read_sym()
			IF sym = sym_range THEN
				read_sym()
				IF sym <> sym_string THEN
					fatal("expected string after range separator")
				END
				##f[s] = f[s] + ".." + StringToLiteral(s)
				f[s] = f[s] + ".." + s
				read_sym()
			END
		ELSE
			RETURN NIL
		END
		RETURN f
	END


	FUNCTION read_termin(): Node
	VAR f, p2, p: Node
	BEGIN
		(* Il primo fattore e' obbligatorio: *)
		p = read_factor()
		IF p = NIL THEN
			fatal("expected factor")
		END
		(* Se c'e' un secondo fattore, allora e' un prodotto: *)
		f = read_factor()
		IF f = NIL THEN
			RETURN p
		END
		p2[type] = prod
		p2[a][0] = p
		p2[a][1] = f
		p = p2
		(* Vedi se ci sono altri fattori: *)
		LOOP
			f = read_factor()
			IF f = NIL THEN
				RETURN p
			END
			p[a][count(p[a])] = f
		END
	END

VAR e, s: Node

BEGIN
	e = read_termin()
	IF sym = sym_vbar THEN
		s[type] = or
		s[a][0] = e
		e = s
	END
	WHILE sym = sym_vbar DO
		read_sym()
		e[a][count(e[a])] = read_termin()
	END
	RETURN e
END


FUNCTION read_decl(): Decl
VAR d: Decl
BEGIN
	IF sym <> sym_name THEN
		fatal("expected rule name")
	END
	d[id] = add_id(s, TRUE)
	read_sym()
	IF sym <> sym_eq THEN
		fatal("expected `='")
	END
	read_sym()
	d[n] = read_expr()
	IF sym <> sym_semicolon THEN
		err("unexpected symbol -- check possible missing `;' in the rule above")
	END
	read_sym()
	RETURN d
END


FUNCTION read_decls(fn: STRING)
VAR
	index: INTEGER
	decl: Decl
BEGIN
	TRY Open(fd, fn, "r") END
	line_n = 1
	line_pos = 0
	TRY line = io.ReadLine(fd) END
	c = NIL
	IF print_source THEN
		PrintLineSource()
	END
	read_ch()
	read_sym()
	WHILE sym <> sym_eof DO
		decl = read_decl()
		inc(index, 1)
		decl[id][index] = index
		decls[count(decls)] = decl
	END
END


FUNCTION sort_decls(decls: ARRAY OF Decl)
VAR i, j: INTEGER  d: Decl
BEGIN
	FOR i=0 TO count(decls)-2 DO
		FOR j=i+1 TO count(decls)-1 DO
			IF decls[i][id][name] > decls[j][id][name] THEN
				d = decls[i]
				decls[i] = decls[j]
				decls[j] = d
			END
		END
		decls[i][id][index] = i+1
	END
END


FUNCTION print_as_text(decls: ARRAY OF Decl)

	FUNCTION print_expr(d: Node)
	VAR i, index: INTEGER
	BEGIN
		IF d[type] = box THEN
			print_text(" " + d[id][name])
			IF do_print_index THEN
				index = d[id][index]
				IF index = 0 THEN
					print("_?")
				ELSE
					print("_" + itos(index))
				END
			END
		ELSIF d[type] = roundbox THEN
			print(" " + d[s])
		ELSIF d[type] = repeat THEN
			print(" {")
			print_expr(d[n])
			print(" }")
		ELSIF d[type] = opt THEN
			print(" [")
			print_expr(d[n])
			print(" ]")
		ELSIF d[type] = prod THEN
			FOR i=0 TO count(d[a])-1 DO
				IF d[a][i][type] = or THEN
					print(" (")
					print_expr(d[a][i])
					print(" )")
				ELSE
					print_expr(d[a][i])
				END
			END
		ELSIF d[type] = or THEN
			FOR i=0 TO count(d[a])-1 DO
				IF i > 0 THEN
					print(" |")
				END
				print_expr(d[a][i])
			END
		ELSE
			error("internal error (1)\n")
			exit(1)
		END
	END

VAR i: INTEGER  d: Decl
BEGIN
	FOR i=0 TO count(decls)-1 DO
		IF do_print_index THEN
			print(itos(i+1) + ". ")
		END
		d = decls[i]
		print(d[id][name])
		print(" = ")
		print_expr(d[n])
		print_text(" ;\n")
	END
END


FUNCTION print_as_html(decls: ARRAY OF Decl)

	FUNCTION print_expr(d: Node)
	VAR i, index: INTEGER
	BEGIN
		IF d[type] = box THEN
			print(" <i>" + d[id][name] + "</i>")
			IF do_print_index THEN
				index = d[id][index]
				IF index = 0 THEN
					print("<sub>?</sub>")
				ELSE
					print("<sub>" + itos(index) + "</sub>")
				END
			END
		ELSIF d[type] = roundbox THEN
			print(" <code><b>" + String2HTML(d[s]) + "</b></code>")
		ELSIF d[type] = repeat THEN
			print(" {")
			print_expr(d[n])
			print(" }")
		ELSIF d[type] = opt THEN
			print(" [")
			print_expr(d[n])
			print(" ]")
		ELSIF d[type] = prod THEN
			FOR i=0 TO count(d[a])-1 DO
				IF d[a][i][type] = or THEN
					print(" (")
					print_expr(d[a][i])
					print(" )")
				ELSE
					print_expr(d[a][i])
				END
			END
		ELSIF d[type] = or THEN
			FOR i=0 TO count(d[a])-1 DO
				IF i > 0 THEN
					print(" |")
				END
				print_expr(d[a][i])
			END
		ELSE
			error("internal error (1)\n")
			exit(1)
		END
	END

VAR i: INTEGER  d: Decl
BEGIN
	FOR i=0 TO count(decls)-1 DO
		IF do_print_index THEN
			print(itos(i+1) + ". ")
		END
		d = decls[i]
		print(d[id][name])
		print(" = ")
		print_expr(d[n])
		print(" ;<br>\n")
	END
END


FUNCTION Help()
BEGIN
	print(
"bnf_chk " + VERSION + " - EBNF syntax checker\n" +
"Usage: bnf_chk {OPTION} file\n" +
"\n" +
"Available options:\n" +
"\t--report       Report unused/undeclared ids\n" +
"\t--sort         Sort declarations\n" +
"\t--print-text   Print decls in ASCII\n" +
"\t--print-html   Print decls in HTML\n" +
"\t--print-index  Print index numbers of decls\n" +
"\t--no-print-file-name Do not print file name in error msgs\n" +
"\t--print-source Print source with line numbers\n" +
"\t--print-context  Print error context\n" +
"\t--help         Displays this help\n" +
"\n")
END


FUNCTION Version()
BEGIN
	print_text("bnf_chk " + VERSION + " - EBNF syntax checker\n")
	print_text("Copyright 2006 by icosaedro.it di Umberto Salsi <salsi@icosaedro.it>\n")
	print_text("More info: http://www.icosaedro.it/bnf_chk/index.html\n")
END


VAR
	args: ARRAY OF STRING
	i: INTEGER

BEGIN
	(* Default preferences: *)
	tab_size = 8
	do_sort = FALSE
	do_report = FALSE
	do_print_text = FALSE
	do_print_html = FALSE
	print_file_name = TRUE
	print_source = FALSE
	print_context = FALSE

	args = get_args()
	FOR i=1 TO count(args)-1 DO
		IF args[i] = "--help" THEN
			Help()
		ELSIF args[i] = "--version" THEN
			Version()
		ELSIF args[i] = "--sort" THEN
			do_sort = TRUE
		ELSIF args[i] = "--report" THEN
			do_report = TRUE
		ELSIF args[i] = "--print-text" THEN
			do_print_text = TRUE
		ELSIF args[i] = "--print-html" THEN
			do_print_html = TRUE
		ELSIF args[i] = "--print-index" THEN
			do_print_index = TRUE
		ELSIF args[i] = "--no-print-file-name" THEN
			print_file_name = FALSE
		ELSIF args[i] = "--print-source" THEN
			print_source = TRUE
		ELSIF args[i] = "--print-context" THEN
			print_context = TRUE
		ELSIF args[i][0] = "-" THEN
			error("ERROR: unknown option " + args[i] + "\n")
			exit(1)
		ELSE
			IF fn <> NIL THEN
				error("ERROR: parameter " + args[i]
				+ ": file name already set to " + fn + "\n")
				error("Please type \"bnf_chk --help\" for a complete list of the available options\n")
				exit(1)
			END
			fn = args[i]
		END
	END
	IF fn = NIL THEN
		error("ERROR: required file name\n")
		exit(1)
	END
	ids = NIL
	decls = NIL
	read_decls(fn)
	IF do_report     THEN ids_report()      END
	IF do_sort       THEN sort_decls(decls) END
	IF do_print_text THEN print_as_text(decls) END
	IF do_print_html THEN print_as_html(decls) END
END

(* THE END *)
